__version__ = '0.4.2'

# "Boil some water, lots of it!"
# "Group the porcupines in pairs"

# needed for inheritance and adding in folders
from OFS.Folder import Folder
from OFS.Image import Image
from OFS.DTMLMethod import DTMLMethod
from Products.ZSQLMethods.SQL import SQL

# usual cut and paste imports
from Globals import DTMLFile
from Globals import MessageDialog
from Persistence import Persistent

from Acquisition import Implicit
import OFS.SimpleItem
import AccessControl.Role
from AccessControl import getSecurityManager

# usual imports
import string
import os
import Globals
import sys
import re

# Z SQL Method support
from Shared.DC.ZRDB.Aqueduct import decodestring, parse
try: from IOBTree import Bucket
except: Bucket=lambda:{}

# dtml files for adding forms
manage_addSQLPollForm = DTMLFile('dtml/addpoll', globals())
manage_addSQLPollQuestionForm = DTMLFile('dtml/addquestion', globals())
manage_addSQLPollAnswerForm = DTMLFile('dtml/addanswer', globals())

def manage_addSQLPoll(self, id, connection_id, dbtype, default_skin='gel_red', title='', REQUEST=None):
    ''' Adds a new SQLPoll object '''
    obj = SQLPoll() # standard folder constructor
    obj.id = id
    obj.title = title
    obj.default_skin = default_skin
    obj._addRole('Poll Manager')
    self._setObject(id, obj)
    
    obj.initialise(connection_id, dbtype) # folder creation stuff
    self.manage_permission('Vote', roles=['Anonymous','Manager'], acquire=1)
    if REQUEST is not None:
        return self.manage_main(self, REQUEST)

def manage_addSQLPollQuestion(self, id, title='', REQUEST=None):
    ''' Adds a new object '''
    obj = SQLPollQuestion() # standard folder constructor
    obj.id = id
    obj.title = title

    self._setObject(id, obj)
    obj.initialise(create_table=self.create_table, add_new_row=self.add_new_row) # folder creation stuff
    if REQUEST is not None:
        return self.manage_main(self, REQUEST)

def manage_addSQLPollAnswer(self, title='', REQUEST=None):
    ''' Adds a new object '''
    id = str(self.get_answer_number()) # finds the id for next answer
    obj = SQLPollAnswer(id, title) # this isnt a folder, I'll happily make an __init__ and call it normally
    self._setObject(id, obj)
    obj.initialise(self) # object creation stuff
    if REQUEST is not None:
        return self.manage_main(self, REQUEST)

# SQLPoll is a container for the rest of the poll so that
# the HTML methods will be visible to users so they can edit it.
class SQLPoll(Folder, Implicit):
    meta_type = 'Polls Holder (SQL)'

    isPrincipiaFolderish = 1
    security = AccessControl.ClassSecurityInfo()

    security.declareObjectProtected('View')
    security.declareProtected('View', '__call__', 'call')
    # hmm, not sure about View security and methods. still need look into this.
    # Ed C. Note:  There are errors with view permissions in /manage

    # weird Zope 2.3 manage_options hack
    _m = list(Folder.manage_options)
    # dominique: changed the view action to display the list of polls.
    _m[1] = {'label':'View','action':'index_html'}
    _m[2] = {'label':'Change Skin','action':'manage_editSkin'}
    manage_options = (_m)

    ### management view methods
    security.declareProtected('View management screens', 'initialise', 'filtered_meta_types', 'manage_editSkin') 

    def initialise(self, connection_id, dbtype):
        ''' adds in DTML Methods, Images and Z SQL Methods from filesystem '''
        for file in ['index_html', 'vote_big_html', 'vote_small_html', 'results_html', 'add_new_vote']:
            self._put_file_in(file)
	
	self.manage_addFolder('images', 'Poll graphics')
	for imagedir in os.listdir(Globals.package_home(globals()) + os.sep + 'images'):
            self._put_imagedir_in(imagedir)
	for zsql in os.listdir(Globals.package_home(globals()) + os.sep + 'sql' + os.sep + dbtype):
	    self._put_zsql_in(zsql, connection_id, dbtype)
        self.root = string.join(self.getPhysicalPath(), '/')
    
    manage_editSkin = DTMLFile('dtml/editpollskin', globals())

    def filtered_meta_types(self):
        '''
        returns the objects we want from the tuple
        so people can only add in Questions into a Polls Holder
        '''
        l = []
        types = Folder.filtered_meta_types(self)
        for t in types:
            # dominique: prevent to add answers AND poll holders inside a poll holder
            if t['name'] not in ['Poll Answer (SQL)','Polls Holder (SQL)']:
                l.append(t)
        return l

    ### Change skin methods
    security.declareProtected('Change Skin', 'change_default_skin')
    
    def change_default_skin(self, skin):
        self.default_skin = skin
        if REQUEST is not None:
            return MessageDialog( title='Edited', message='<strong>%s</strong> has a new default skin.' % self.id, action ='./manage_main', )
	
    ### private methods
    def _read_file(self, filename):
        ''' reads file from filesystem '''
        file = open(filename, 'rb')
        data = file.read()
        file.close()
        return data

    def _put_file_in(self, file):
        ''' reads then puts file in ZODB '''
        data = self._read_file(Globals.package_home(globals()) + os.sep + file)
        obj = DTMLMethod(data, __name__ = file)
	obj._proxy_roles = ('Manager',)
        self._setObject(file, obj)  
   
    def _put_imagedir_in(self, folder):
        ''' reads then puts image in ZODB '''

        self.images.manage_addFolder(folder,'Images') # create Folder
        # dominique: fixed bad indenting in that block. Problems showed up with Windows NT
        for file in os.listdir(Globals.package_home(globals()) + os.sep + 'images' + os.sep + folder):
            m = re.match('^(.*)\.(\w+)$', file)
            name, ext = m.groups()
            if ext in ['gif', 'jpg', 'jpeg', 'png']:
                data = self._read_file(Globals.package_home(globals()) + os.sep + 'images' + os.sep + folder + os.sep + file)
                self.images[folder].manage_addImage(file, data, title='Poll Image')
            if ext == 'dtml':
                data = self._read_file(Globals.package_home(globals()) + os.sep + 'images' + os.sep + folder + os.sep + file)
                self.images[folder].manage_addDTMLMethod(name, title='Poll Method', file=data)

	
    def _put_zsql_in(self, file, connection_id, dbtype):
	''' reads then puts ZSQLMethod in ZODB '''
	f = open(Globals.package_home(globals()) + os.sep + 'sql' + os.sep + dbtype + os.sep + file, 'r')
        data = f.read()
        f.close()
 	# the z sql methods use the FTP-style of putting params at the top
	# this parsing method was copied from ./lib/python/Shared/DC/ZRDB/DA.py's PUT function
	obj = SQL(id=file, title='Poll SQL Method',
		  connection_id=connection_id,
		  arguments='',
		  template='')
	m = re.match('<params>(.*)</params>\s*\n', data, re.I | re.S) # wish this was Perl :)
	if m:
	    obj.arguments_src = m.group(1)
	    obj._arg=parse(obj.arguments_src)
	    data = data[m.end():]
	else:
	    data = 'Waaaaa! could not parse the sql text properly\n\n' + data
	template = data
	obj.src = template
	obj.template=t=obj.template_class(template)
	t.cook()
	obj._v_cache={}, Bucket()
	obj.manage_permission(permission_to_manage='Use Database Methods',
	       roles=['Manager', 'Poll Manager'], acquire=0) # Manager access only
	self._setObject(file, obj)
    
# SQLPollQuestion is the object for the poll question, its a container
# so it can have the answers in it.
class SQLPollQuestion(Folder, Implicit):
    meta_type = 'Poll Question (SQL)'
    isPrincipiaFolderish = 1

    security = AccessControl.ClassSecurityInfo()

    # copy and paste would screw up the SQL database, if people had two answers
    # pointing to the same column on the SQL
    dontAllowCopyAndPaste = 1
    property_extensible_schema__ = 0

    # weird Zope 2.3 manage_options hack
    _m = list(Folder.manage_options)
    _m[1] = {'label':'View','action':'vote_big_html'}
    _m[2] = {'label':'Properties','action':'manage_editForm'}
    manage_options = (_m)

    ### View management methods
    security.declareProtected('View management screens', 'initialise', 'filtered_meta_types', 
                              'manage_edit',  'manage_editForm', 'get_answer_number')
    
    manage_editForm = DTMLFile('dtml/SQLPollQuestionEdit', globals()) # edit form

    def initialise(self, create_table, add_new_row):
        ''' start it up '''
        self.open = 1
	self.tablename = 'p_q_' + str(self.id)
	# create an SQL table
	create_table(tablename=self.tablename)
	# add_new_row(tablename=self.tablename)

    def filtered_meta_types(self):
        '''
        returns the objects we want from the tuple
        so people can only add in Answers into a Question
        '''
        l = []
        types = Folder.filtered_meta_types(self)
        for t in types:
            if t['name'] in ['Poll Answer (SQL)']:
                l.append(t)
        return l

    def manage_edit(self, title, open=0, REQUEST=None):
        ''' edits the poll '''
        self.title=title
        self.open=open
        if REQUEST is not None:
            return MessageDialog( title='Edited', message='<strong>%s</strong> has been edited.' % self.id, action ='./manage_main', )

    def get_answer_number(self):
        ''' gets an id for the next answer '''
        k = 0
        n = "Answer%s" % k
        for i in self.objectIds(['Poll Answer (SQL)']):
            k = k + 1
            n = "Answer%s" % k
        return n
    
    ### Voting methods
    security.declareProtected('Vote', 'add_vote')
    
    def add_vote(self, answer, poll_identifier, REQUEST=None):
        ''' count vote '''
        #print REQUEST,
        #obj = self
        if not self.open:
        #    if REQUEST is not None:
            return MessageDialog(title='Error', message='This poll is not open.', action ='../.' )
        #    else:
        #        return None
        if self.repeat_voter(poll_identifier):
            return MessageDialog(title='Error', message='You may only vote once.', action ='../.' )
        else:
            #print answer
            #print poll_identifier
            answerobj = getattr(self, answer)
            #print answerobj
            try:
                answerobj.add_vote(poll_identifier)
	    #    if REQUEST is not None:
	    #	    REQUEST.RESPONSE.redirect('./results_html')
                return MessageDialog(title='Vote Entered', message='Your vote was addded to the database', action = './results_html')
            except:
	    #    if REQUEST is not None:
	         return MessageDialog(title='Error', message='Sorry an internal error occured, your vote was not counted.<!-- %s, %s-->' % (sys.exc_type, sys.exc_value), action ='./results_html')

    ### Voting Security Methods

    def repeat_voter(self, poll_identifier):
        ''' Is the voter a duplicate?  Returns 1 if yes, 0 if no '''
        for row in self.aq_parent.select_repeat_voter(tablename=self.tablename, poll_identifier=poll_identifier): 
            count = row[0]
        #print "repeat voter: %i" % count
        return count

    ### 'enhanced' view methods
    security.declareProtected('Access contents information', 'total_votes')

    def total_votes(self):
        ''' get the number of total votes '''
        for row in self.aq_parent.select_total_votes(tablename=self.tablename):
	    total = row[0]
	if not total:
	    total = 0.1
	if total < 1:  # a kludge to avoid division by zero errors
	    total = 0.1
        return total

    ### need to protect these methods still
    def manage_beforeDelete(self, item, container):
        ''' clean up '''
        self._delete()

    ### private methods
    def _delete(self):
	''' delete entire question '''
	try:
	    self.aq_parent.drop_table(tablename=self.tablename)
	    return 1
	except:
	    return 0


# SQLPollAnswer is the actual answer, it's a simple item with data in a column in an SQL table
class SQLPollAnswer(OFS.SimpleItem.Item,
                    Persistent,
		    Implicit,
		    AccessControl.Role.RoleManager):
    meta_type = 'Poll Answer (SQL)'

    manage_options = ( ( {'label':'Edit', 'action':'manage_editForm'}, ) + (OFS.SimpleItem.Item.manage_options))

    security = AccessControl.ClassSecurityInfo()
    security.declareProtected('View management screens', 'initialise', 'manage_edit',
                              'manage_beforeDelete', 'manage_editForm', 'reset')
    
    def __init__(self, id, title='', REQUEST=None):
        ''' init the object '''
        self.id = id
        self.title = title
        self.answerattr = 'a_' + str(self.id)

        if REQUEST is not None:
            return MessageDialog( title='Added', message='Question has been added.', action ='./manage_main', )

    def initialise(self, parent):
        ''' change answerattr '''
        parent.aq_parent.add_attr(tablename=parent.tablename, attrname=self.answerattr)
        manage_editForm = DTMLFile('dtml/SQLPollAnswerEdit', globals())

    def manage_edit(self, title, reset=0, REQUEST=None):
        ''' edits the answer '''
        self.title = title
        if REQUEST is not None:
            # dominique: return to the question holding the edited answer.
            return MessageDialog(title = 'Edited', message = "Answer edited", action = '../manage_main',)

    def manage_beforeDelete(self, item, container):
        ''' clean up '''
        self._delete()

    def reset(self, REQUEST=None):
        ''' resets the question by deleting the values '''
        # dominique: added support for reset
        value = self.read()
        self.aq_parent.aq_parent.write(tablename=self.aq_parent.tablename, attrname=self.answerattr, total=0)
        self.aq_parent.aq_parent.write(tablename=self.aq_parent.tablename, attrname='total_votes', total=int(self.aq_parent.total_votes() - value))
        if REQUEST is not None:
            return MessageDialog(title = 'Edited', message = "Votes reset", action = './manage_editForm',)

    ### 'enhanched' view methods
    security.declareProtected('Access contents information', 'votes', 'read')

    def votes(self):
        ''' number of votes '''
        return self.read()

    def read(self):
        ''' read column '''
        try:
	    for row in self.aq_parent.aq_parent.read(tablename=self.aq_parent.tablename, attrname=self.answerattr):
		value = row[0]
	    if not value: value = 0
        except:
            value = 0 # error reading column
        return value

    ### Vote methods
    security.declareProtected('Vote', 'add_vote')

    def add_vote(self, poll_identifier):
        ''' increase '''
        #print "Answer object add_vote"
        return self._write(poll_identifier)

    ### private methods
    def _delete(self):
        ''' delete all answers '''
        try:
	    # this isn't going to work with PostgreSQL, since it doesn't have a drop attribute feature.
	    # John sez "that's not part of the SQL standard, and it should be up to the developer
	    # to implement the renaming of the table, creation of a new table, and then copying
	    # over the data from the old table and finally droping the table."
	    # Which seems like a lot of work to me ...
	    self.aq_parent.drop_attr(tablename=self.tablename,attrname=self.answerattr)
            return 1
        except:
            return 0

    def _write(self, poll_identifier):
        ''' increment vote by one '''
        ## Ed C Note:  The vars are correct, It can't find this write SQL method?

        #print "answer object _write"
        #value = int(self.read())
	#if not value:
	#    value = 0
        #print self.aq_parent.tablename
        #print self.answerattr
        #print poll_identifier
        #print self.aq_parent
        #print self.aq_parent.aq_parent
        #print self.aq_parent.aq_parent.aq_parent
        #print self.aq_parent.aq_parent.aq_parent.aq_parent
	self.aq_parent.aq_parent.write(tablename=self.aq_parent.tablename,
	      attrname=self.answerattr, poll_identifier=poll_identifier)
	# increase total_votes as well
	#self.aq_parent.aq_parent.write(tablename=self.aq_parent.tablename,
	#      attrname='total_votes', total=int(self.aq_parent.total_votes() + 1))
        return int(1)
